 // Autor: Hélio Sousa
 // Trabalho de Computação Gráfica - Efólio A - 2022-2023
 // Implementação do algoritmo do ponto médio para desenhar linhas raster
 // PARTE I- Implementação do algoritmo do ponto médio (2 valores):

// variaveis globais
let declive = false;    // variavel booleana para verificar se o declive é superior a 1
let simetrico = false;  // variavel booleana para verificar o sinal do declive

/**
* Caso geral do Algoritmo do ponto médido como descrito em (Foley et al.,1995,p.78)
* O caso geral, é aplicado ao primeiro octante (0º a 45º) ou (E a NE)
* Se o segmento de reta estiver noutro octante, é executada uma transformação geométrica de redução
* ao primeiro octante e, posteriormente, uma transformação inversa de volta ao octante original.
* @param {*} P extremo inferior inicial
* @param {*} Q extremo superior final
* @returns devolve um array com os pontos do segmento de reta
*/
function lineMP(P = { x: x0, y: y0 }, Q = { x: x1, y: y1 }) {
  let newP = {x: P.x, y: P.y};         // guarda os valores iniciais de P
  let newQ = {x: Q.x, y: Q.y};         // guarda os valores iniciais de Q
  let listaPontos = [];                // lista que vai receber os pontos
  reduceToFirstOctant(newP, newQ);  // transformação de redução ao primeiro octante
  let dx = newQ.x - newP.x;            // calcular dx novos pontos reduzidos ao Primeiro Octante
  let dy = newQ.y - newP.y;            // Calculo de dy novos pontos reduzidos ao Primeiro Octante
  let d = 2 * dy - dx;                 // Fator de decisão inicial
  let incrE = 2 * dy;                  // Incremento usado para mover para E
  let incrNE = 2 * (dy - dx);          // Incremento usado para mover para NE
  let x = newP.x;                      // Inicialização do x
  let y = newP.y;                      // Inicialização do y
  listaPontos.push(newP);              // Insere o ponto inicial no array de pontos
  while (x < newQ.x) {                 // Enquanto x for menor que x1
    if (d <= 0) {                      // Se d for menor ou igual a zero
      d+=incrE;                        // Incrementa d com o valor de incrE
      x++;                             // Incrementa x
    } else {                           // Se não
      d+=incrNE;                       // Incrementa d com o valor de incrNE
      x++;                             // Incrementa x
      y++;                             // Incrementa y
    }
    listaPontos.push({x:x ,y: y});     // insere o ponto no array de pontos
  }
  // transformação inversa -> devolver os pontos ao octante original
  if(declive) {                                 // Se o declive for verdadeiro
      for(let i=0; i<listaPontos.length; i++) { // Para cada ponto na lista de pontos
          let aux = listaPontos[i].x;           // trocar os valores  das coordenadas x e y
          listaPontos[i].x = listaPontos[i].y;  // 
          listaPontos[i].y = aux;               // 
      }
  }
  if(simetrico) {                               // Se simetrico for verdadeiro
      for(let i=0; i<listaPontos.length; i++) { // Para cada ponto na lista de pontos
          listaPontos[i].y = -listaPontos[i].y; // inverter o sinal da coordenada y
      }
  }
  // Note-se que, como é indiferente traçar um segmento de reta do primeiro para o segundo
  // extremo ou do segundo para o primeiro, uma eventual troca da ordem dos extremos
  // não necessita de ser invertida na transformação inversa (Brisson,2013,p.12)
  
  return listaPontos;     // Retorna a lista de pontos
}

/**
 * Redução ao primeiro octante como definida em (Brisson,2013,pp.12-13)
 * @param {*} newP extremo inicial passado por referência
 * @param {*} newQ extremo final passado por referência
 * Reduz os pontos newP e newQ ao primeiro octante
 * Se newP e newQ estiverem no primeiro octante, não faz nada
 */
function reduceToFirstOctant(newP, newQ) {
    declive = false;                   // 1 definir dois valores lógicos, declive e simétrico
    simetrico = false;                 // e inicializá-los a falso
    let dx = newQ.x - newP.x;          // 2.1 calcular dx inicial 
    let dy = newQ.y - newP.y;          // 2.2 calcular dy final
    let dx_x_dy = dx*dy;               // 3 testar o produto de dx por dy,
    if (dx_x_dy < 0) {                 // 3.1 se negativo
        newP.y = -newP.y;              // 3.2 substituir o valor de newP.y pelo simétrico
        newQ.y = -newQ.y;              // 3.3 substituir o valor de newQ.y pelo simétrico
        dy = -dy;                      // 3.4 substituir o valor de dy pelo simétrico
        simetrico = true;              // 3.4 atribuir a simétrico verdadeiro
    }
    if (Math.abs(dy) > Math.abs(dx)) { // 4. testar se o valor absoluto de dx é maior que dx, se verdadeiro
        let aux = newP.x;              // 4.1 trocar os valores  das coordenadas x e y de newP
        newP.x = newP.y;               // 
        newP.y = aux;                  // 
        aux = newQ.x;                  // 4.2 trocar os valores  das coordenadas x e y de newQ
        newQ.x = newQ.y;               //
        newQ.y = aux;                  //
        aux = dx;                      // 4.3 trocar os valores de dx e dy
        dx = dy;                       //
        dy = aux                       //
        declive = true;                // 4.4 atribuir a declive verdadeiro
    }
    if (newP.x > newQ.x) {             // 5. testar se o valor de newP.x é maior que o valor de newQ.x, se verdadeiro
        let aux = newP.x;              // 5.1 trocar os valores de newP.x por newQ.x
        newP.x = newQ.x;               //
        newQ.x = aux;                  //
        aux = newP.y;                  // 5.2 trocar os valores de newP.y por newQ.y
        newP.y = newQ.y;               //
        newQ.y = aux;                  //
    }
}

export default lineMP;       // exporta lineMP