// Autor: Hélio Sousa
// Trabalho de Computação Gráfica - Efólio A - 2022-2023
// Implementação do algoritmo do ponto médio para desenhar linhas raster
// PARTE II - Interface Gráfico (2 valores)

// Importação de módulos
import * as THREE from 'https://cdn.skypack.dev/three@0.132.2';
import { OrbitControls } from 'https://cdn.skypack.dev/three@0.132.2/examples/jsm/controls/OrbitControls.js'
import lineMP from '../lineMP.mjs'

// Inicialização da função principal
document.addEventListener('DOMContentLoaded', start);

// variaveis globais
let P;                        // primeiro ponto selecionado
let Q;                        // segundo ponto selecionado
const sizePixel = 1;          // tamanho do pixel(e tile) irão manter-se sempre a 1
const sizeRasterDisplay = 10; // tamanho da grelha inicial de pixels
let rows = sizeRasterDisplay; // número de colunas do display raster para respeitar o enunciado, é um tabuleiro de lados iguais
let cols = sizeRasterDisplay; // número de linhas do display raster para respeitar o enunciado, é um tabuleiro de lados iguais

const colorTile = "yellow";     // consultado em https://en.wikipedia.org/wiki/X11_color_names
const colorSq1 = 0xff8966;      // consultado em https://imagecolorpicker.com/en -> cor (0,0)
const colorSq2 = 0x8d88b5;      // consultado em https://imagecolorpicker.com/en -> cor (0,1)
const colorIntersected = "red"; // consultado em https://en.wikipedia.org/wiki/X11_color_names

const raycaster = new THREE.Raycaster(); // cria uma instancia de raycaster https://threejs.org/docs/#api/en/core/Raycaster
raycaster.layers.set(1);                 // raycaster apenas interseta objetos na layer 1 https://threejs.org/docs/#api/en/core/Raycaster.layers
let mouse = new THREE.Vector2();         // cria um vector2, automaticamente inicializado com (0,0) -> https://threejs.org/docs/#api/en/math/Vector2
let firstXPressed = false;               // variavel auxiliar para saber se o primeiro ponto já foi selecionado

// Definir parametros da cena
// renderer 
let renderer = new THREE.WebGLRenderer();               // cria uma instancia de renderer
renderer.setSize(window.innerWidth,window.innerHeight); // define o tamanho do renderer (viewport)
document.body.appendChild(renderer.domElement);         // O canvas onde o renderizador irá desenhar o seu output
// scene 
let scene = new THREE.Scene();                          // cria uma instância de cena
scene.background = new THREE.Color( 0xffffff );         // muda a cor do background da cena para branco
// camera
let camera = new THREE.PerspectiveCamera(90, 
    window.innerWidth / window.innerHeight, 1, 1000);   // cria uma instância de camera
scene.add(camera);                                      // adiciona a camera à cena

// orbitControls -> https://threejs.org/docs/#examples/en/controls/OrbitControls
let orbitControls = new OrbitControls(camera, renderer.domElement); // ) //chamada dos orbit controls para o utilizador interagir com a cena
camera.position.set(0, 0, sizeRasterDisplay+1); // colocar a camera numa posição inicial que permita ver todo o tabuleiro
orbitControls.update();
camera.lookAt( 0, 0, 0 );                       // faz a camera olhar para o centro da scene

// definir luz ambiente
let lightAmbient = new THREE.AmbientLight(0xffffff, 1); // branco = 0xffffff
scene.add(lightAmbient);

// FUNÇAO MAIN + ATUALIZAR FRAMES
/**
 * Inicia a animação, função principal chamada no DOMContentLoaded
 */
function start() {
    displayRaster()                // desenhar o chão (floor)
    axesXYHelper();                // Desenha os eixos X e Y
    requestAnimationFrame(animate); // chamada da função animate
}

/**
 * Atualiza a cena com as entradas do utilizador
 */
function animate() {
    renderer.render(scene, camera); // renderiza a cena
    requestAnimationFrame(animate);  // chamada recursiva da função animate

}

// EVENTOS
// EventListener que devolve a posição do rato quando este se move em coordenadas normalizadas entre -1 e +1
document.body.addEventListener('mousemove', ev=> {
    mouse.x = (ev.clientX / window.innerWidth) * 2 - 1;
    document.getElementById('mouseX').textContent = mouse.x.toFixed(2);
    document.getElementById('mouseX_A').textContent = Math.round(mouse.x*(sizeRasterDisplay+1)*(window.innerWidth/window.innerHeight)); // para que a progressão seja de -sizeRasterDisplay a +sizeRasterDisplay é necessário multiplicar pelo aspect ratio do viewport largura/altura
    mouse.y = - (ev.clientY / window.innerHeight) * 2 + 1;
    document.getElementById('mouseY').textContent = mouse.y.toFixed(2);
    document.getElementById('mouseY_A').textContent = Math.round(mouse.y*(sizeRasterDisplay+1)); // sizeRasterDisplay é o número de linhas do display raster, +1 que é a linha do zero
});
// get mouse position -> https://codepen.io/dcode-software/pen/exbKLv
window.addEventListener('mousemove', ev=> {
    document.getElementById('x').textContent = ev.x;
    document.getElementById('y').textContent = ev.y;
    document.getElementById('innerWidth').textContent = window.innerWidth;
    document.getElementById('innerHeight').textContent = window.innerHeight;
});

// EventListener para teclas pressionadas
document.body.addEventListener('keydown', ev=> {
    if (ev.code === "Backspace") {                                         // se a tecla pressionada for Backspace -> https://www.toptal.com/developers/keycode/for/backspace
        resetScene();                                                      // chama a função resetScene
    }
    if (ev.code === "KeyX") {                                              // se a tecla pressionada for 'x' ou 'X' -> https://www.toptal.com/developers/keycode/for/x
        raycaster.setFromCamera(mouse, camera);                            // inicializa o raycaster com as posições do rato em relação á camara
        const intersects = raycaster.intersectObjects(scene.children);         // guarda o pixel que eventualmente poderá ser intersetado pelo Raycaster
        if(firstXPressed) {                                                // se o firstXPressed está true, então este é segundo X e será guardado no ponto Q (final)
            if(intersects.length > 0) {                                    // se o array intersects tiver elementos (ou seja, se o raycaster intercetou algum objeto)
                intersects[0].object.material.color.set(colorIntersected); // altera a cor do pixel intercetado para vermelho
                Q = {x: intersects[0].object.position.x,                   // o segundo ponto intercetado pelo ray, guarda a posição x do pixel intercetado (Q.x)
                    y: intersects[0].object.position.y};                   // o segundo ponto intercetado pelo ray, guarda a posição y do pixel intercetado (Q.y)
                firstXPressed = false;                                     // retornar o firstXPressed a false, para que o próximo X pressionado seja guardado num novo ponto P (inicial)
                drawTilesMP();                                         // Chama a função que irá desenhar a inserção da linha raster na scene
                drawRealLine();                                            // Chama a função que irá desenhar a linha verdadeira.
            }
        } else {                                                           // se o firstXPressed está false, então este é primeiro X e será guardado no ponto P (inicial)
            if(intersects.length > 0) {                                    // se o array intersects tiver elementos (ou seja, se o raycaster intercetou algum objeto)
                intersects[0].object.material.color.set(colorIntersected)  // altera a cor do pixel intercetado para vermelho
                P = {x: intersects[0].object.position.x,                   // o primeiro ponto intercetado pelo ray, guarda a posição x do pixel intercetado (P.x)
                    y: intersects[0].object.position.y};                   // o primeiro ponto intercetado pelo ray, guarda a posição x do pixel intercetado (P.y)
                firstXPressed = true;                                      // o primeiro ponto já foi selecionado, o próximo X pressionado será guardado no ponto Q (final)
            }
        }
    }
});

// FUNÇÕES AUXILIARES
/**
 * Criação dos eixos cartesianos x e y positivos
 * Nota seria possivel criar de forma mais simples com o AxesHelper, mas acrescentaria o eixo z e a alternativa
 * seria customizar o codigo fonte do AxesHelper, o que não é o objetivo deste trabalho
 */
function  axesXYHelper() {
    const pointsAxisX = [];                                                        // array que irá conter os pontos do eixo X
    pointsAxisX.push( new THREE.Vector3( 0, 0, 0.01 ) );                           // ponto inicial do eixo X
    pointsAxisX.push( new THREE.Vector3((sizeRasterDisplay+(sizePixel)/2),0,0.01));// ponto final do eixo X, é o tamanho do tabuleiro + metade do tamanho de um pixel, porque o eixo X começa no meio do pixel
    let geometryAxisX = new THREE.BufferGeometry().setFromPoints( pointsAxisX);    // criação da geometria do eixo X
    let materialAxisX = new THREE.LineBasicMaterial( { color: 0x0000ff  } );       // criação do material do eixo X
    let axisX = new THREE.Line( geometryAxisX, materialAxisX);                     // criação do eixo X
    
    const pointsAxisY = [];                                                        // array que irá conter os pontos do eixo Y
    pointsAxisY.push( new THREE.Vector3( 0, 0, 0.01 ) );                           // ponto inicial do eixo Y
    pointsAxisY.push( new THREE.Vector3(0,(sizeRasterDisplay+(sizePixel/2)),0.01));// ponto final do eixo Y, é o tamanho do tabuleiro + metade do tamanho de um pixel, porque o eixo Y começa no meio do pixel
    let geometryAxisY = new THREE.BufferGeometry().setFromPoints( pointsAxisY );   // criação da geometria do eixo Y
    let materialAxisY = new THREE.LineBasicMaterial( { color: 0xff0000} );         // criação do material do eixo Y
    let axisY = new THREE.Line( geometryAxisY, materialAxisY );                    // criação do eixo Y
    axisX.layers.enable(2);                                                        // ativa a camada 2 para o eixo X
    axisY.layers.enable(2);                                                        // ativa a camada 2 para o eixo Y
    scene.add( axisX, axisY );                                                     // adicionar os eixos X e Y à scene
}

/** 
 * Função que vai recomeçar a scene
 */
 function resetScene() {
    scene.remove.apply(scene, scene.children);      // Remove todos os filhos da scene (apaga tudo)
    displayRaster();                                // Chama a função que vai desenhar o display raster
    axesXYHelper();                                 // Reconstroi os eixos cartesianos
    camera.position.set(0, 0, sizeRasterDisplay+1); // Volta a colocar a camera na posição inicial
    camera.lookAt(0, 0, 0);                         // Volta a colocar a camera a olhar para o centro da scene
}

/** 
 * Função que cria um pixel para ser colocado no display raster
 * @param {*} side          // tamanho do lado do pixel
 * @param {*} pixelColor    // cor do pixel
 * @param {*} x             // posição x do pixel
 * @param {*} y             // posição y do pixel
 * @returns                 // retorna o pixel criado
 */
 function createPixel(side, pixelColor, x, y) {
    const geometry = new THREE.PlaneGeometry(side, side);               // criação da geometria de um plano com o tamanho unitário, sizel pixe = 1
    const material = new THREE.MeshBasicMaterial({color: pixelColor,    // criação do material do pixel 
        side: THREE.DoubleSide});                                       // para o objeto ficar sempre visivel independentemente da posição da camera
    let pixel = new THREE.Mesh(geometry, material);                     // criação do pixel
    pixel.position.set(x,y,0);                                          // coloca o pixel na posição x,y do display raster, com z=0
    return pixel;                                                       // retorna o pixel
}

/**
 * Função que vai criar o display raster, recebe como arg o tamanho da grelha, 
 * o tamanho de cada pixel (lado do quadrado) a cor1 e a cor2
 * adaptado de https://physevangertis.medium.com/loops-in-javascript-make-checkerboards-cf9c17308072
 */
 function displayRaster() {
    let color;
    for (let x = -rows; x <= rows; x++) {     //ciclo em x de -tamanho da grelha até +tamanho da grelha
        for (let y = -cols; y <= cols; y++) { //ciclo em y de -tamanho da grelha até +tamanho da grelha
            if((y % 2 == 0 && x %2 == 0) ||  // se ambos x e y forem pares
                (x % 2 != 0 && y % 2 != 0)){ // ou ambos forem impares
                color = colorSq1;             // atribui a cor 1
            }
            else {                            // se x for par e y for impar ou vice versa
                color = colorSq2;             // atribui a cor 2
            }
            let pixel = new createPixel(sizePixel, color, x, y); // coloca em pixel o pixel(mesh) criado nas posições (x,y) com a cor apropriada
            pixel.layers.enable(1);                              // ativa a camada 1 para o pixel
            scene.add(pixel);                                    // adiciona o pixel à scene
        }
    }
}

/** 
 * Função que cria a linha verdadeira
 */
 function drawRealLine() {                                                       // P e Q são os pontos extremos da linha
    const pointsRealLine = [];                                                   // array que irá conter os pontos da linha verdadeira
    pointsRealLine.push( new THREE.Vector3( P.x, P.y, sizePixel/4 ));            // ponto inicial da linha verdadeira
    pointsRealLine.push( new THREE.Vector3( Q.x, Q.y, sizePixel/4 ));            // ponto final da linha verdadeira
    let geometryLine = new THREE.BufferGeometry().setFromPoints(pointsRealLine); // criação da geometria da linha verdadeira
    let materialLine = new THREE.LineBasicMaterial( { color: "black"} );         // criação do material da linha verdadeira
    let realLine = new THREE.Line(geometryLine, materialLine);                   // criação da linha verdadeira
    scene.add(realLine);                                                         // adicionar a linha verdadeira à scene
}

/**
 * Função que cria o Tile para compor a lineMP
 * @param {*} pixelPosition posição do pixel devolvido pelo algoritmo lineMP
 * @returns ladrilho a desenhar
 */
 function createTile(pixelPosition) {
    const tileHeight = (sizePixel) / 4;                                        // Inicializa a altura do tile com 1/4 do tamanho do pixel
    const geometry = new THREE.BoxGeometry(sizePixel, sizePixel, tileHeight);  // cria uma nova boxgeometry, de lado sizePixel, altura sizePixel e profundidade tileHeight
    const material = new THREE.MeshBasicMaterial({                             // cria o material
        opacity: 0.5,                                                          // define a opacidade em 0.5 (transparente)
        transparent: true,                                                     // ativa a transparência transparente
        color: colorTile,                                                      // define a cor dada por colorPixel
        side: THREE.DoubleSide                                                 // com double side
    });
    let tile = new THREE.Mesh(geometry, material);  // cria a Mesh do ladrilho  
    tile.position.x = pixelPosition.x;              // fica na posição x recebida do cálculo de lineMP
    tile.position.y = pixelPosition.y;              // fica na posição y recebida do cálculo de lineMP
    tile.position.z = tileHeight / 2;               // divide por dois para ficar ao nivel do display raster, e não metade por baixo e a outra metade por cima
    return tile;
}

/**
 * Função que cria a linha MP
 */
 function drawTilesMP() {                   // P e Q são os pontos extremos da linha
    let pixelsPositions = [];               // array que irá conter os pontos da linha MP
    pixelsPositions = lineMP(P, Q);         // recebe os pontos da linha MP
    const tilesMP = [];                     // array para receber os tiles da linha
    pixelsPositions.every(pixelPosition => tilesMP.push(createTile(pixelPosition))); // cria um tile para cada ponto da linha MP
    tilesMP.every(tile => scene.add(tile)); // adiciona cada tile à scene
}